package net.cattaka.slidepuzzle;

import java.sql.SQLException;
import java.util.List;

import net.cattaka.slidepuzzle.db.DbHandler;
import net.cattaka.slidepuzzle.dto.InData;
import net.cattaka.slidepuzzle.dto.Score;
import net.cattaka.slidepuzzle.solver.Solver;
import net.cattaka.slidepuzzle.solver.World;
import net.cattaka.slidepuzzle.util.StringUtil;

public class SolverRunner {
	public static void run(boolean onlyUnSolved, boolean retryAnswered, int solverType, int stepLimit, int currentLimit, int historyLimit, int begin, int end) throws SQLException {
		DbHandler dbHandler = new DbHandler();
		dbHandler.initialize();
		for (int i=begin;i<=end;i++) {
			System.out.println("Solver is running : " + i);
			InData inData = dbHandler.findIndata(i);
			if (inData == null) {
				continue;
			}
			if (!retryAnswered) {
				if (onlyUnSolved) {
					List<Score> oldScores = dbHandler.findScore(i);
					if (oldScores.size() > 0) {
						System.out.println("    skipped : " + i);
						continue;
					}
				} else {
					Score oldScore = dbHandler.findScore(i,solverType);
					if (oldScore != null) {
						System.out.println("    skipped : " + i);
						continue;
					}
				}
			}
			World world = World.createWorld(inData.getBi(), inData.getWi(), inData.getHi());
			Solver solver = new Solver();
			solver.initialize(world, solverType);
			World result = solver.solve(stepLimit, currentLimit, historyLimit);
			if (result != null) {
				System.out.println(StringUtil.conv(result.getExPath()));
				Score score = new Score();
				score.setId(inData.getId());
				score.setSolverType(solver.getSolverType());
				score.setAnswer(StringUtil.conv(result.getExPath()));
				score.recalc();
				dbHandler.registerScore(score);
			}
		}
	}
	
	public static void main(String[] args) throws Exception {
		boolean onlyUnSolved = false;
		boolean retryAnswered = false;
		int solverType = 0;
		int stepLimit = 200;
		int currentLimit = 100000;
		int historyLimit = 1000000;
		int begin = 1;
		int end = 5000;
		
		int mode = 0;
		for (String arg : args) {
			switch (mode) {
			case 0:
				mode = ("-s".equals(arg)) ? 1 : mode;
				mode = ("-l".equals(arg)) ? 2 : mode;
				mode = ("-c".equals(arg)) ? 3 : mode;
				mode = ("-h".equals(arg)) ? 4 : mode;
				mode = ("-b".equals(arg)) ? 5 : mode;
				mode = ("-e".equals(arg)) ? 6 : mode;
				if ("-r".equals(arg)) {
					retryAnswered = true;
				}
				if ("-i".equals(arg)) {
					solverType = -1;
				}
				if ("-O".equals(arg)) {
					onlyUnSolved = true;
				}
				break;
			case 1:
				solverType = Integer.parseInt(arg);
				mode = 0;
				break;
			case 2:
				stepLimit = Integer.parseInt(arg);
				mode = 0;
				break;
			case 3:
				currentLimit = Integer.parseInt(arg);
				mode = 0;
				break;
			case 4:
				historyLimit = Integer.parseInt(arg);
				mode = 0;
				break;
			case 5:
				begin = Integer.parseInt(arg);
				mode = 0;
				break;
			case 6:
				end = Integer.parseInt(arg);
				mode = 0;
				break;
			}
		}
		if (solverType == 0) {
			System.out.println("usage");
			System.out.println(" -s SolverType : 1,2(required)");
			System.out.println(" -l StepLimit : num(default:"+stepLimit+")");
			System.out.println(" -c CurrentLimit : num(default:"+currentLimit+")");
			System.out.println(" -h HistoryLimit : num(default:"+historyLimit+")");
			System.out.println(" -b Begin : num(default:"+begin+")");
			System.out.println(" -e End : num(default:"+end+")");
			System.out.println(" -r retryAnswered : (default:"+retryAnswered+")");
			System.out.println(" -O onlyUnSolved : (default:"+onlyUnSolved+")");
		} else if (solverType == -1) {
			InDataLoader.load();
		} else {
			System.out.println("SolverType : " + solverType);
			System.out.println("StepLimit : " + stepLimit);
			System.out.println("CurrentLimit : " + currentLimit);
			System.out.println("HistoryLimit : " + historyLimit);
			run(onlyUnSolved, retryAnswered, solverType, stepLimit, currentLimit, historyLimit, begin, end);
		}
	}
}
