package net.cattaka.slidepuzzle;

import java.sql.SQLException;
import java.util.List;

import net.cattaka.slidepuzzle.db.DbHandler;
import net.cattaka.slidepuzzle.dto.InData;
import net.cattaka.slidepuzzle.dto.Score;
import net.cattaka.slidepuzzle.solver.UltimetSolver;
import net.cattaka.slidepuzzle.solver.World;
import net.cattaka.slidepuzzle.util.StringUtil;

public class UltimetSolverRunner {
	public static void run(boolean retryAnswered, boolean onlyUnSolved, int solverType, int begin, int end) throws SQLException {
		DbHandler dbHandler = new DbHandler();
		dbHandler.initialize();
		List<InData> inDatas;
		if (onlyUnSolved) {
			inDatas = dbHandler.findIndataByUnsolved(begin, end);
		} else {
			inDatas = dbHandler.findAllIndata(begin, end);
		}
		for (InData inData : inDatas) {
			System.out.println("Solver is running : " + inData.getId());
			if (inData == null) {
				continue;
			}
			if (!retryAnswered) {
				Score oldScore = dbHandler.findScore(inData.getId(),solverType);
				if (oldScore != null) {
					System.out.println("    skipped : " + inData.getId());
					continue;
				}
			}
			World world = World.createWorld(inData.getBi(), inData.getWi(), inData.getHi());
			World result = null;
			try {
				UltimetSolver solver1 = new UltimetSolver();
				solver1.initialiez(world, true);
				World result1 = solver1.solve();

				UltimetSolver solver2 = new UltimetSolver();
				solver2.initialiez(world, true);
				World result2 = solver2.solve();
				
				if (result1.getExPath().length < result2.getExPath().length) {
					result = result1;
				} else {
					result = result2;
				}
			} catch (SolverException e) {
				e.printStackTrace();
			}
			if (result != null) {
				System.out.println(StringUtil.conv(result.getExPath()));
				Score score = new Score();
				score.setId(inData.getId());
				score.setSolverType(3);
				score.setAnswer(StringUtil.conv(result.getExPath()));
				score.recalc();
				dbHandler.registerScore(score);
			}
		}
	}
	
	public static void main(String[] args) throws Exception {
		boolean onlyUnSolved = false;
		boolean retryAnswered = false;
		int solverType = 4;
		int begin = 1;
		int end = 5000;
		
		int mode = 0;
		for (String arg : args) {
			switch (mode) {
			case 0:
				mode = ("-s".equals(arg)) ? 1 : mode;
				mode = ("-b".equals(arg)) ? 5 : mode;
				mode = ("-e".equals(arg)) ? 6 : mode;
				if ("-r".equals(arg)) {
					retryAnswered = true;
				}
				if ("-i".equals(arg)) {
					solverType = -1;
				}
				if ("-O".equals(arg)) {
					onlyUnSolved = true;
				}
				break;
			case 1:
				solverType = Integer.parseInt(arg);
				mode = 0;
				break;
			case 5:
				begin = Integer.parseInt(arg);
				mode = 0;
				break;
			case 6:
				end = Integer.parseInt(arg);
				mode = 0;
				break;
			}
		}
		if (solverType == 0) {
			System.out.println("usage");
			System.out.println(" -s SolverType : 1,2(required)");
			System.out.println(" -b Begin : num(default:"+begin+")");
			System.out.println(" -e End : num(default:"+end+")");
			System.out.println(" -r retryAnswered : (default:"+retryAnswered+")");
			System.out.println(" -O onlyUnSolved : (default:"+onlyUnSolved+")");
		} else if (solverType == -1) {
			InDataLoader.load();
		} else {
			System.out.println("SolverType : " + solverType);
			run(retryAnswered, onlyUnSolved, solverType, begin, end);
		}
	}
}
