package net.cattaka.slidepuzzle.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import net.cattaka.slidepuzzle.dto.InData;
import net.cattaka.slidepuzzle.dto.Score;

public class DbHandler {
	public static final String DB_FILE = "slidepuzlle.db";
	
	public DbHandler() {
	}
	
	public void initialize() throws SQLException {
		try {
			Class.forName("org.sqlite.JDBC");
		} catch (ClassNotFoundException e) {
			throw new SQLException(e);
		}
		Connection conn = null;
	    try {
			conn = getConnection();
			conn.setAutoCommit(false);
		    Statement stmt = conn.createStatement();
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS IN_DATA (" +
					"SEQ INTEGER PRIMARY KEY," +
					"ID INTEGER UNIQUE," +
					"WI INTEGER," +
					"HI INTEGER," +
					"BI TEXT)"
				);
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS SCORE (" +
					"SEQ INTEGER PRIMARY KEY," +
					"ID INTEGER," +
					"SOLVER_TYPE INTEGER," +
					"ANSWER TEXT," +
					"COUNT INTEGER," +
					"COUNT_L INTEGER," +
					"COUNT_R INTEGER," +
					"COUNT_U INTEGER," +
					"COUNT_D INTEGER," +
					"UNIQUE(ID,SOLVER_TYPE))"
				);
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS COUNTER(SEQ INTEGER PRIMERY KEY)");
			for (int i=1;i<=5000;i++) {
				stmt.executeUpdate(String.format("INSERT INTO COUNTER(SEQ) VALUES (%d)", i));
			}
			stmt.close();
			conn.commit();
	    } finally {
	    	if (conn != null) {
	    		conn.close();
	    	}
	    }
	}
	
	public Connection getConnection() throws SQLException {
	    return DriverManager.getConnection("jdbc:sqlite:" + DB_FILE);
	}
	
	public void registerInData(Connection conn, InData inData) throws SQLException {
		try {
			try {
				PreparedStatement stmt = conn.prepareStatement("INSERT INTO IN_DATA(ID,WI,HI,BI) VALUES (?,?,?,?)");
			    stmt.setInt(1, inData.getId());
			    stmt.setInt(2, inData.getWi());
			    stmt.setInt(3, inData.getHi());
			    stmt.setString(4, inData.getBi());
			    stmt.executeUpdate();
			    stmt.close();
			} catch (SQLException e) {
				PreparedStatement stmt = conn.prepareStatement("UPDATE IN_DATA SET WI=?,HI=?,BI=? WHERE ID=?");
			    stmt.setInt(1, inData.getWi());
			    stmt.setInt(2, inData.getHi());
			    stmt.setString(3, inData.getBi());
			    stmt.setInt(4, inData.getId());
			    stmt.executeUpdate();
			    stmt.close();
			}
		} finally {
		}
	}
	public void registerScore(Score score) throws SQLException {
		Connection conn = null;
		try {
			conn = getConnection();
			try {
				PreparedStatement stmt = conn.prepareStatement("INSERT INTO SCORE(ID,SOLVER_TYPE,ANSWER,COUNT,COUNT_L,COUNT_R,COUNT_U,COUNT_D) VALUES (?,?,?,?,?,?,?,?)");
			    stmt.setInt(1, score.getId());
			    stmt.setInt(2, score.getSolverType());
			    stmt.setString(3, score.getAnswer());
			    stmt.setInt(4, score.getCount());
			    stmt.setInt(5, score.getCountL());
			    stmt.setInt(6, score.getCountR());
			    stmt.setInt(7, score.getCountU());
			    stmt.setInt(8, score.getCountD());
			    stmt.executeUpdate();
			    stmt.close();
			} catch (SQLException e) {
				PreparedStatement stmt = conn.prepareStatement("UPDATE SCORE SET ANSWER=?,COUNT=?,COUNT_L=?,COUNT_R=?,COUNT_U=?,COUNT_D=? WHERE ID=? AND SOLVER_TYPE=?");
			    stmt.setString(1, score.getAnswer());
			    stmt.setInt(2, score.getCount());
			    stmt.setInt(3, score.getCountL());
			    stmt.setInt(4, score.getCountR());
			    stmt.setInt(5, score.getCountU());
			    stmt.setInt(6, score.getCountD());
			    stmt.setInt(7, score.getId());
			    stmt.setInt(8, score.getSolverType());
			    stmt.executeUpdate();
			}
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
	}
	public InData findIndata(int id) throws SQLException {
		InData inData = null;
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery("SELECT ID,WI,HI,BI FROM IN_DATA WHERE ID=" + id);
			if (rs.next()) {
				inData = new InData();
				inData.setId(rs.getInt(1));
				inData.setWi(rs.getInt(2));
				inData.setHi(rs.getInt(3));
				inData.setBi(rs.getString(4));
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return inData;
	}
	public List<InData> findAllIndata(int begin, int end) throws SQLException {
		List<InData> result = new ArrayList<InData>();
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(String.format("SELECT ID,WI,HI,BI FROM IN_DATA WHERE ID BETWEEN %d AND %d ORDER BY ID", begin, end));
			while (rs.next()) {
				InData inData = new InData();
				inData.setId(rs.getInt(1));
				inData.setWi(rs.getInt(2));
				inData.setHi(rs.getInt(3));
				inData.setBi(rs.getString(4));
				result.add(inData);
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return result;
	}
	public List<InData> findIndataByUnsolved(int begin, int end) throws SQLException {
		List<InData> result = new ArrayList<InData>();
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(String.format("SELECT ID,WI,HI,BI FROM IN_DATA WHERE ID NOT IN (SELECT ID FROM SCORE) AND ID BETWEEN %d AND %d ORDER BY ID", begin, end));
			while (rs.next()) {
				InData inData = new InData();
				inData.setId(rs.getInt(1));
				inData.setWi(rs.getInt(2));
				inData.setHi(rs.getInt(3));
				inData.setBi(rs.getString(4));
				result.add(inData);
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return result;
	}
	public Score findScore(int id, int solverType) throws SQLException {
		Score score = null;
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(String.format("SELECT ID,SOLVER_TYPE,ANSWER FROM SCORE WHERE ID=%d AND SOLVER_TYPE=%d",id,solverType));
			if (rs.next()) {
				score = new Score();
				score.setId(rs.getInt(1));
				score.setSolverType(rs.getInt(2));
				score.setAnswer(rs.getString(3));
				score.recalc();
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return score;
	}
	public List<Score> findScore(int id) throws SQLException {
		List<Score> results = new ArrayList<Score>();
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(String.format("SELECT ID,SOLVER_TYPE,ANSWER FROM SCORE WHERE ID=%d ORDER BY SOLVER_TYPE", id));
			while (rs.next()) {
				Score score = new Score();
				score.setId(rs.getInt(1));
				score.setSolverType(rs.getInt(2));
				score.setAnswer(rs.getString(3));
				score.recalc();
				results.add(score);
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return results;
	}
	public List<Score> findAllScore() throws SQLException {
		List<Score> results = new ArrayList<Score>();
		Connection conn = null;
		try {
			conn = getConnection();
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(String.format("SELECT ID,SOLVER_TYPE,ANSWER FROM SCORE ORDER BY ID,SOLVER_TYPE"));
			while (rs.next()) {
				Score score = new Score();
				score.setId(rs.getInt(1));
				score.setSolverType(rs.getInt(2));
				score.setAnswer(rs.getString(3));
				score.recalc();
				results.add(score);
			}
			
			rs.close();
			stmt.close();
		} finally {
			if (conn != null) {
				conn.close();
			}
		}
		return results;
	}
}
