package net.cattaka.slidepuzzle.solver;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

public class SolveStageType1 implements ISolveStage {
	private byte wx;
	private byte wy;

	public SolveStageType1() {
	}

	public void initialiez(World world) {
		this.wx = world.getWx();
		this.wy = world.getWy();
	}

	@Override
	public int calcScore(World world) {
		for (int i=0;i<world.getTile().length;i++) {
			byte v = world.getTile()[i];
			if (v == -1 || v == 0 || (v-1) == i) {
				continue;
			}
			world.setExScore(0);
			return 0;
		}
		world.setExScore(1);
		return 1;
	}

	@Override
	public int getMinScore() {
		return 0;
	}

	@Override
	public int getMaxScore() {
		return 1;
	}
	
	@Override
	public int getSolverType() {
		return 1;
	}

	public void excludeUncertainlyCurrentSet(Set<World> worlds, int limit) {
		class WorldBundle implements Comparable<WorldBundle> {
			int score;
			World world;
			public int compareTo(WorldBundle o){
				return this.score - o.score;
			}
		}
		ArrayList<WorldBundle> bundles = new ArrayList<WorldBundle>(worlds.size());
		for (World world : worlds) {
			WorldBundle bundle = new WorldBundle();
			bundle.world = world;
			bundle.score = 0;
			for (int i=0;i<world.getTile().length;i++) {
				byte v = world.getTile()[i];
				if (v == -1 || v == 0) {
					// none
				} else {
					bundle.score += 12 - (Math.abs(((v-1)%wx)-(i%wx)) + Math.abs(((v-1)/wx)-(i/wx)));
				}
			}
			bundles.add(bundle);
		}
		Collections.sort(bundles);
		for (int i=0;i<bundles.size()-limit;i++) {
			worlds.remove(bundles.get(i).world);
		}
	}
	@Override
	public void excludeUncertainlyHistoryMap(Map<World, World> worlds, int limit) {
		worlds.clear();
	}

	@Override
	public boolean isCutByScore() {
		return true;
	}

	public boolean isStepByScore() {
		return true;
	}
	
	public boolean isStepPosition(int x, int y) {
		return true;
	}
}
