package net.cattaka.slidepuzzle.solver;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.cattaka.slidepuzzle.util.StringUtil;

public class SolveStageType2 implements ISolveStage {
	private int wx;
	private int wy;
	private boolean[] recordTargets;
	private byte[] weights;
	private byte[][] distances;
	private int maxWeight = 0;
	
	@Override
	public int calcScore(World world) {
		int score = 0;
		for (int i = 0 ;i<world.getTile().length;i++) {
			int v = world.getTile()[i];
			if (v > 0) {
				int dist = distances[v - 1][i];
				//int weight = (int) Math.pow((wx * wy), weights[v - 1]);
				int weight = (int) (wx * wy) * weights[v - 1];
				score -= weight * dist;
			}
		}
		world.setExScore(score);
		return score;
	}

	@Override
	public void excludeUncertainlyCurrentSet(Set<World> worlds, int limit) {
		final Comparator<World> comparator = new Comparator<World>() {
			@Override
			public int compare(World o1, World o2) {
				return o2.getExScore() - o1.getExScore();
			}
		};
		ArrayList<World> worldList = new ArrayList<World>(worlds);
		Collections.sort(worldList, comparator);
		for (int i=limit;i<worldList.size();i++) {
			worlds.remove(worldList.get(i));
		}
	}

	@Override
	public void excludeUncertainlyHistoryMap(Map<World, World> worlds, int limit) {
//		worlds.clear();
		final Comparator<World> comparator = new Comparator<World>() {
			@Override
			public int compare(World o1, World o2) {
				return o2.getExScore() - o1.getExScore();
			}
		};
		ArrayList<World> worldList = new ArrayList<World>(worlds.keySet());
		Collections.sort(worldList, comparator);
		for (int i=limit/2;i<worldList.size();i++) {
			worlds.remove(worldList.get(i));
		}
	}

	@Override
	public int getMinScore() {
		return Integer.MIN_VALUE;
	}
	@Override
	public int getMaxScore() {
		// このソルバーはマイナスに振れるので0が最大
		return 0;
	}

	@Override
	public int getSolverType() {
		return 2;
	}

	@Override
	public void initialiez(World world) {
		{
			int len = world.getTile().length;
			this.wx = world.getWx();
			this.wy = world.getWy();
			this.recordTargets = new boolean[len];
			this.weights = new byte[len];
			this.distances = new byte[len][len];
			{	// 壁を抜き出す
				for (int i=0;i<this.weights.length;i++) {
					if (world.getTile()[i] == -1) {
						weights[i] = -1;
					}
				}
			}
		}
		{	// 各セルの分岐点(方向が３以上のものと右下)にチェックをつける
			for (int x=0;x<wx;x++) {
				for (int y=0;y<wy;y++) {
					byte w = weights[x+wx*y];
					if (w == 0) {
						int rn = calcRouteNum(weights, wx, wy, x, y, 1);
						if (rn >= 3) {
							recordTargets[x+wx*y] = true;
						}
					}
				}
			}
			recordTargets[recordTargets.length - 1] = true;
			//System.out.println(StringUtil.convBoolean(recordTarget, wx, wy));
		}
		{	// 各セルの重要度を算出する(Weight)
			{	// 各セルの重要度を算出する
				maxWeight = 0;
				boolean goFlag;
				do	{
					maxWeight++;
					goFlag = false;
					for (int x=0;x<wx;x++) {
						for (int y=0;y<wy;y++) {
							int idx = x+wx*y;
							if (idx == weights.length - 1) {
								continue;
							}
							byte w = weights[idx];
							if (w == 0) {
								int rn = calcRouteNum(weights, wx, wy, x, y, maxWeight);
								if (rn <= 2) {
									weights[idx] = (byte) maxWeight;
									goFlag = true;
								}
							}
						}
					}
					for (int i=1;i<maxWeight;i++) {
						for (int x = 0; x < wx-2; x++) {
							for (int y = 0; y < wy-2; y++) {
								byte v = weights[x + wx * y];
								if (v>0) {
									if (weights[(x+1) + wx * y] > v && weights[(x+2) + wx * y] == v) {
										weights[(x+1) + wx * y] = v;
									}
									if (weights[x + wx * (y+1)] > v && weights[x + wx * (y+2)] == v) {
										weights[x + wx * (y+1)] = v;
									}
								}
							}
						}
					}
				} while(goFlag);
				for (int i=0;i<this.weights.length;i++) {
					if (weights[i] != -1) {
						weights[i] = (byte)(maxWeight - weights[i]);
					}
				}
			}
			{	// 右下は0が来るので重要度を強制的に落とす
				int endWeight = weights[weights.length - 1];
				while (endWeight > 1) {
					List<Integer> result = new ArrayList<Integer>();
					boolean[] picked = new boolean[weights.length];
					pickConnectedGroup(result, weights, picked, endWeight, wx, wy, wx-1, wy-1);
					endWeight--;
					for (int idx : result) {
						weights[idx] = (byte) endWeight;
					}
				}
			}
		}
		{	// 座標1→座標2への最短距離のマップを作成する
			for (int x=0;x<wx;x++) {
				for (int y=0;y<wy;y++) {
					int idx = x + wx*y;
					for (int i=0;i<distances[idx].length;i++) {
						distances[idx][i] = -1;
					}
					calcDistanceMap(weights, distances[idx], 0, wx, wy, x, y);
				}
			}
			//System.out.println(StringUtil.convWeight(distance[7], wx, wy));
		}
	}
	/**
	 * 指定された座標と同じWeightのセルのindexを再帰的にピックする
	 */
	private static void pickConnectedGroup(List<Integer> result, byte[] weight, boolean[] picked, int level, int wx, int wy, int x, int y) {
		int idx = x + y * wx;
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return;
		}
		if (weight[idx] == level) {
			if (picked[idx]) {
				return;
			} else {
				picked[idx] = true;
				result.add(idx);
				pickConnectedGroup(result, weight, picked, level, wx, wy, x + 1, y);
				pickConnectedGroup(result, weight, picked, level, wx, wy, x, y + 1);
				pickConnectedGroup(result, weight, picked, level, wx, wy, x, y - 1);
				pickConnectedGroup(result, weight, picked, level, wx, wy, x - 1, y);
			}
		}
	}
	/**
	 * 指定されたx,yを起点とした各座標への計算された距離のマップを作成する
	 */
	private static void calcDistanceMap(byte[] weight, byte[] out, int dist, int wx, int wy, int x, int y) {
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return;
		}
		int idx = x + wx * y;
		if (weight[idx] == -1) {
			return;
		}
		if (out[idx] == -1 || out[idx] > dist) {
			out[idx] = (byte) dist;
			int nextDist = dist + 1;
			calcDistanceMap(weight, out, nextDist, wx, wy, x-1, y);
			calcDistanceMap(weight, out, nextDist, wx, wy, x+1, y);
			calcDistanceMap(weight, out, nextDist, wx, wy, x, y-1);
			calcDistanceMap(weight, out, nextDist, wx, wy, x, y+1);
		}
	}
	private static int calcRouteNum(byte[] weight, int wx, int wy, int x, int y, int level) {
		int n = 0;
		n += (!isWall(weight, wx, wy, x-1, y, level)) ? 1 : 0;
		n += (!isWall(weight, wx, wy, x+1, y, level)) ? 1 : 0;
		n += (!isWall(weight, wx, wy, x, y-1, level)) ? 1 : 0;
		n += (!isWall(weight, wx, wy, x, y+1, level)) ? 1 : 0;
		return n;
	}
	private static boolean isWall(byte[] weight, int wx, int wy, int x, int y, int level) {
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return true;
		}
		byte w = weight[x+wx*y];
		return (w == -1 || (w > 0 && level > w));
	}

	@Override
	public boolean isCutByScore() {
		return false;
	}
	
	@Override
	public String toString() {
		return StringUtil.convByte(weights, wx, wy);
	}

	public boolean isStepByScore() {
		return false;
	}
	public boolean isStepPosition(int x, int y) {
		return recordTargets[x + wx*y];
	}

	public byte[][] getDistances() {
		return distances;
	}
}
