package net.cattaka.slidepuzzle.solver;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;


public class Solver {
		private World firstWorld;
		private ISolveStage solveStage;
		private byte wx;
		private byte wy;

		public Solver() {
		}

		public void initialize(World world, int solverType) {
			this.wx = world.getWx();
			this.wy = world.getWy();
			this.firstWorld = world;
			if (solverType == 1) {
				this.solveStage = new SolveStageType1();
			} else if (solverType == 2) {
				this.solveStage = new SolveStageType2();
			} else {
				throw new RuntimeException("Unknown solver type : " + solverType);
			}
			this.solveStage.initialiez(world);
		}

		public World solve(int stepLimit, int currentLimit, int historyLimit) {
			byte[] inputs = { 'U', 'L', 'R', 'D'};
			HashMap<World, World> historyWorldMap = new HashMap<World, World>();
			HashSet<World> nextSet = new HashSet<World>();
			HashSet<World> currSet = new HashSet<World>();

			solveStage.calcScore(firstWorld);
			nextSet.add(firstWorld);
			historyWorldMap.put(firstWorld, firstWorld);
			int count = 0;
			int maxScore = solveStage.getMinScore();
			World lastWorld = null;
			do {
				{
					HashSet<World> t = nextSet;
					nextSet = currSet;
					currSet = t;
					nextSet.clear();
				}
				if (currSet.size() == 0) {
					// ありえないががが
					return null;
				}
				if (currSet.size() > currentLimit) {
					// 仕方なく、、
					solveStage.excludeUncertainlyCurrentSet(currSet, currentLimit);
				}
				if (historyWorldMap.size() > historyLimit) {
					solveStage.excludeUncertainlyHistoryMap(historyWorldMap, historyLimit);
					// 仕方なく、、
				}
				for (World world : currSet) {
					lastWorld = world;
					for (byte input : inputs) {
						World nextWorld = stepWorld(world, solveStage, world.getExPath(), input);
						if (nextWorld != null) {
							World historyWorld = historyWorldMap.get(nextWorld);
							if (historyWorld == null) {
								//historyWorldMap.put(nextWorld, nextWorld);
								if (solveStage.isCutByScore()) {
									if (maxScore < nextWorld.getExScore()) {
										maxScore = nextWorld.getExScore();
										// スコアが増えたので一旦クリアする
										historyWorldMap.clear();
										for (Iterator<World> it = nextSet.iterator();it.hasNext();) {
											World nnWorld = it.next();
											if (nnWorld.getExScore() < maxScore) {
												it.remove();
											}
										}
									}
									if (maxScore <= nextWorld.getExScore()) {
										// スコアが減っている場合は次の検索はしない
										historyWorldMap.put(nextWorld, nextWorld);
										nextSet.add(nextWorld);
									}
								} else {
									if (maxScore < nextWorld.getExScore()) {
										maxScore = nextWorld.getExScore();
									}
									historyWorldMap.put(nextWorld, nextWorld);
									nextSet.add(nextWorld);
								}
								
								if (nextWorld.getExScore() == solveStage.getMaxScore()) {
									return nextWorld;
								}
							} else if (nextWorld.getExPath().length < historyWorld.getExPath().length) {
								historyWorldMap.put(nextWorld, nextWorld);
							}
						}
					}
				}
				count++;
				System.out.println(String.format("%d : %d : %d : %d",
						(count), historyWorldMap.size(), currSet.size(),
						maxScore));
				System.out.println(lastWorld);
			} while (count < stepLimit);
			
			return null;
		}

		public World stepWorld(World world, ISolveStage solveStage, byte[] path,
				byte input) {
			Byte c = input;
			World nextWorld = world;

			int score = solveStage.calcScore(nextWorld);
			do {
				nextWorld = nextWorld.getNext(wx, wy, c);
				if (nextWorld == null) {
					break;
				}
				c = nextWorld.getAnotherWay(c, nextWorld.getCursorX(), nextWorld.getCursorY(), wx, wy);
				solveStage.calcScore(nextWorld);
				if (c == null) {
					break;
				}
				if (solveStage.isStepByScore()) {
					if (nextWorld.getExScore() != score) {
						break;
					}
				} else {
					if (solveStage.isStepPosition(nextWorld.getCursorX(), nextWorld.getCursorY())) {
						break;
					}
				}
			} while (nextWorld != null && c != null);

			return nextWorld;
		}
		
		public ISolveStage getSolveStage() {
			return solveStage;
		}

		public int getSolverType() {
			return solveStage.getSolverType();
		}
}
