package net.cattaka.slidepuzzle.solver;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import net.cattaka.slidepuzzle.SolverException;
import net.cattaka.slidepuzzle.util.SolverUtil;
import net.cattaka.slidepuzzle.util.StringUtil;

public class UltimetSolver {
	public static final byte[][] DIR_LIST = new byte[][] {
		{ 0,-1},
		{-1, 0},
		{ 0, 1},
		{ 1, 0},
	};

	public static class PathEntry {
		public int x;
		public int y;
		public List<Byte> path;
		public PathEntry prev;
		public PathEntry(int x, int y, List<Byte> path) {
			super();
			this.x = x;
			this.y = y;
			this.path = path;
		}
		public PathEntry(PathEntry src) {
			super();
			this.x = src.x;
			this.y = src.y;
			this.path = new ArrayList<Byte>(src.path);
		}
		@Override
		public String toString() {
			return StringUtil.conv(path) + String.format("(%d,%d)",x,y);
		}
	}
	public static class SolveStep {
		public int stepType = 0;
		public List<Integer> groupIdxs;
		public SolveStep(int stepType, List<Integer> groupIdxs) {
			super();
			this.stepType = stepType;
			this.groupIdxs = groupIdxs;
		}
		@Override
		public String toString() {
			return "" + stepType + ":" + String.valueOf(groupIdxs);
		}
	}

	private int wx;
	private int wy;
	private byte[] correctAnswer;
	private byte[] weights;
	private List<SolveStep> solveStepList;
	private World firstWorld;
	private PathEntry fakeStartPath;
	
	public void initialiez(World world, boolean ccw) {
		firstWorld = world;
		wx = world.getWx();
		wy = world.getWy();
		weights = new byte[world.getTile().length];
		correctAnswer = new byte[world.getTile().length];
		int count = 0;
		int lastCount = -1;
		for (int i=0;i<world.getTile().length;i++) {
			if (world.getTile()[i] == -1) {
				weights[i] = -1;
				correctAnswer[i] = -1;
			} else {
				weights[i] = 0;
				correctAnswer[i] = (byte)((i != world.getTile().length - 1) ? i+1 : 0);
				count++;
			}
		}
		int x = wx - 1;
		int y = wy - 1;
		int dir = -1;
		{
			// スタート地点の近くに壁がある場合はずらす、それ以外はそのまま
			int[] vs = findNearestSquire(weights, wx, wy, wx-1, wy-1);
			boolean[] mask = new boolean[weights.length];
			x = vs[0]%wx;
			y = vs[0]/wx;
			dir = vs[1];
			List<PathEntry> peList = findMinPath(mask, wx-1, wy-1, x, y);
			if (peList.size() > 1) {
				SolverUtil.swapByte(correctAnswer, wx, wy, wx-1, wy-1, peList.get(peList.size()-1).path);
			}
			fakeStartPath = peList.get(peList.size()-1);
		}
		if (!ccw) {
			dir = (dir <= 0) ? 3 : dir - 1;
		}
		int firstIdx = x + wx*y;
		byte level = 1;
		byte lastLevel = 1;
		byte fakeStartLevel = 1;
		weights[firstIdx] = level;
		boolean detectingFlag = true;
		int detectingStartIdx = -1;
		int detectingStartDir = 0;
		int dirDiff = 0;
		int lastIdx = -1;
		solveStepList = new ArrayList<SolveStep>();
		List<Integer> groupIdxs = new ArrayList<Integer>();
		groupIdxs.add(firstIdx);
		
		int bugLimit = 0;
		while(count > 0 && (bugLimit++ < 10000)) {
			byte nextWeight = -1;
			int nx = 0;
			int ny = 0;
			
			if (firstIdx == x+wx*y && lastIdx != x+wx*y) {
				if (count == lastCount) {
					// 一本道の場所があるのでそこを通り抜ける
					int[] vs = findNearestSquire(weights, wx, wy, x, y);
					int tx = vs[0]%wx;
					int ty = vs[0]/wx;
					boolean[] mask = new boolean[weights.length];
					for (int i=0;i<weights.length;i++) {
						mask[i] = (weights[i] == -1);
					}
					List<PathEntry> peList = findMinPath(mask, x, y, tx, ty);
					PathEntry pe = null;
					int nextDir = -1;
					{
						int i;
						for (i=peList.size()-1;i>=0;i--) {
							PathEntry tPe = peList.get(i);
							byte tw = weights[tPe.x + wx*tPe.y];
							weights[tPe.x + wx*tPe.y] = 0;
							int tDir = checkSquire(weights, wx, wy, tPe.x, tPe.y);
							if (tDir == -1) {
								count--;
								weights[tPe.x + wx*tPe.y] = tw;
								break;
							}
							count++;
							pe = tPe;
							nextDir = tDir;
						}
						for (;i>=0;i--) {
							PathEntry tPe = peList.get(i);
							if (weights[tPe.x + wx*tPe.y] == 0) {
								count--;
								weights[tPe.x + wx*tPe.y] = level;
								groupIdxs.add(tPe.x + wx*tPe.y);
							} else {
								break;
							}
						}
						if (groupIdxs.size() > 0) {
							level++;
							solveStepList.add(new SolveStep(2, groupIdxs));
							groupIdxs = new ArrayList<Integer>();
						}
					}
					x = pe.x;
					y = pe.y;
					dir = nextDir;
					if (!ccw) {
						dir = (dir <= 0) ? 3 : dir - 1;
					}
					weights[x+wx*y] = level;
					groupIdxs.add(x+wx*y);
					
					lastCount = count;
					lastLevel = (byte) (level - 1);
					fakeStartLevel = level;
					weights[x+y*wx] = level;
					detectingFlag = true;
					detectingStartDir = dir;
					detectingStartIdx = -1;
				}
				lastCount = count;
				lastLevel = level;
			}
			
			if (detectingFlag) {
				dir = (dir+2) % 4;
				dirDiff+=2;
				for (int i = 0;i<DIR_LIST.length;i++) {
					if (ccw) {
						dir = (dir <= 0) ? 3 : dir - 1;
					} else {
						dir = (dir+1) % 4;
					}
					dirDiff-=1;
					nx = x + DIR_LIST[dir][0];
					ny = y + DIR_LIST[dir][1];
					nextWeight = getWeight(weights, wx, wy, nx, ny);
					if (nextWeight == -1 || nx + wx*ny == detectingStartIdx) {
						continue;
					} else if (nextWeight == 0) {
						lastIdx = nx + wx*ny;
						groupIdxs.add(nx + wx*ny);
						weights[nx + wx*ny] = level;
						break;
					} else if (dirDiff < 0 || (level > fakeStartLevel && level == nextWeight)) {
						// 不正なパスなので巻き戻す
						for (Integer idx : groupIdxs) {
							weights[idx] = 0;
						}
						groupIdxs.clear();
						lastIdx = detectingStartIdx;
						nx = detectingStartIdx%wx;
						ny = detectingStartIdx/wx;
						detectingFlag = false;
						dir = (detectingStartDir+2)%4;
						break;
					} else {
						count-=groupIdxs.size();
						level++;
						detectingFlag = false;
						int stepType = 1;
						{
							if (groupIdxs.size() == 4) {
								int s = groupIdxs.get(0);
								int e = groupIdxs.get(3);
								if (Math.abs(s-e) == 1 || ((s-e)%wx == 0 && Math.abs((s-e)/wx) == 1)) {
									stepType = 0;
								}
							}
						}
						solveStepList.add(new SolveStep(stepType, groupIdxs));
						groupIdxs = new ArrayList<Integer>();
						break;
					}
				}
			} else {
				dir = (dir+2) % 4;
				for (int i = 0;i<DIR_LIST.length;i++) {
					if (ccw) {
						dir = (dir+1) % 4;
					} else {
						dir = (dir <= 0) ? 3 : dir - 1;
					}
					nx = x + DIR_LIST[dir][0];
					ny = y + DIR_LIST[dir][1];
					nextWeight = getWeight(weights, wx, wy, nx, ny);
					if (nextWeight == -1 || lastLevel <= nextWeight) {
						continue;
					} else if (nextWeight == 0) {
						lastIdx = nx + wx*ny;
						groupIdxs.add(nx + wx*ny);
						weights[nx + wx*ny] = level;
						detectingFlag = true;
						detectingStartIdx = x + wx*y;
						detectingStartDir = dir;
						dirDiff = 0;
						break;
					} else {
						// none
						break;
					}
				}
			}
			x = nx;
			y = ny;
		}
		
		if (bugLimit >= 10000) {
			throw new SolverException("Program error");
		}
		
		Collections.reverse(solveStepList);
	}

	private static byte getWeight(byte[] weights, int wx, int wy, int x, int y) {
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return -1;
		}
		return weights[x + wx*y];
	}
	
	/** 戻り値は{初期idx, 初期dir} */
	private static int[] findNearestSquire(byte[] weights, int wx, int wy, int x, int y) {
		List<int[]> posList = new LinkedList<int[]>();
		posList.add(new int[]{x,y});
		boolean[] checked = new boolean[weights.length];
		while (posList.size() > 0) {
			int[] pos = posList.remove(0);
			int idx = pos[0] + wx * pos[1];
			if (pos[0] < 0 || wx <= pos[0] || pos[1] < 0 || wy <= pos[1]) {
				continue;
			}
			if (checked[idx]) {
				continue;
			}
			if (getWeight(weights, wx, wy, pos[0], pos[1]) == -1) {
				continue;
			}
			checked[idx] = true;
			boolean b11 = (getWeight(weights, wx, wy, pos[0]-1, pos[1]-1) == 0);
			boolean b12 = (getWeight(weights, wx, wy, pos[0]-1, pos[1]) == 0);
			boolean b13 = (getWeight(weights, wx, wy, pos[0]-1, pos[1]+1) == 0);
			boolean b21 = (getWeight(weights, wx, wy, pos[0], pos[1]-1) == 0);
			boolean b22 = (getWeight(weights, wx, wy, pos[0], pos[1]) == 0);
			boolean b23 = (getWeight(weights, wx, wy, pos[0], pos[1]+1) == 0);
			boolean b31 = (getWeight(weights, wx, wy, pos[0]+1, pos[1]-1) == 0);
			boolean b32 = (getWeight(weights, wx, wy, pos[0]+1, pos[1]) == 0);
			boolean b33 = (getWeight(weights, wx, wy, pos[0]+1, pos[1]+1) == 0);
			if (b11 && b12 && b21 && b22) {
				return new int[]{idx, 3};
			}
			if (b21 && b22 && b31 && b32) {
				return new int[]{idx, 2};
			}
			if (b12 && b13 && b22 && b23) {
				return new int[]{idx, 0};
			}
			if (b22 && b23 && b32 && b33) {
				return new int[]{idx, 1};
			}
			posList.add(new int[]{pos[0],pos[1]-1});
			posList.add(new int[]{pos[0]-1,pos[1]});
			posList.add(new int[]{pos[0],pos[1]+1});
			posList.add(new int[]{pos[0]+1,pos[1]});
		}
		throw new SolverException("Program error or bad data.");
	}
	private static int checkSquire(byte[] weights, int wx, int wy, int x, int y) {
		boolean b11 = (getWeight(weights, wx, wy, x-1, y-1) == 0);
		boolean b12 = (getWeight(weights, wx, wy, x-1, y) == 0);
		boolean b13 = (getWeight(weights, wx, wy, x-1, y+1) == 0);
		boolean b21 = (getWeight(weights, wx, wy, x, y-1) == 0);
		boolean b22 = (getWeight(weights, wx, wy, x, y) == 0);
		boolean b23 = (getWeight(weights, wx, wy, x, y+1) == 0);
		boolean b31 = (getWeight(weights, wx, wy, x+1, y-1) == 0);
		boolean b32 = (getWeight(weights, wx, wy, x+1, y) == 0);
		boolean b33 = (getWeight(weights, wx, wy, x+1, y+1) == 0);
		if (b11 && b12 && b21 && b22) {
			return 3;
		}
		if (b21 && b22 && b31 && b32) {
			return 2;
		}
		if (b12 && b13 && b22 && b23) {
			return 0;
		}
		if (b22 && b23 && b32 && b33) {
			return 1;
		}
		return -1;
	}
		
	@Override
	public String toString() {
		return StringUtil.convByte(weights, wx, wy);
	}
	
	public List<PathEntry> findMinPath(boolean[] mask, int x1, int y1, int x2, int y2) {
		List<PathEntry> peList = new LinkedList<PathEntry>();
		boolean[] checked = new boolean[weights.length];
		peList.add(new PathEntry(x1,y1, new ArrayList<Byte>()));
		for (int breaker=0;breaker<1000;breaker++) {
			if (peList.size() == 0) {
				throw new SolverException("Program error or bad data.");
			}
			PathEntry pe = peList.remove(0);
			if (x2 == pe.x && y2 == pe.y) {
				List<PathEntry> result = new ArrayList<PathEntry>();
				do {
					result.add(pe);
					pe = pe.prev;
				} while (pe != null);
				Collections.reverse(result);
				return result;
			}
			if (checked[pe.x + pe.y*wx]) {
				continue;
			} else {
				checked[pe.x + pe.y*wx] = true;
			}
			if (weights[pe.x + pe.y*wx] == -1) {
				continue;
			}
			if (pe.prev != null && mask[pe.x + pe.y*wx]) {
				continue;
			}
			
			byte last = (pe.path.size() > 0) ? pe.path.get(pe.path.size()-1) : 0;
			if (0 < pe.x && last != 'R') {
				PathEntry nextPe = new PathEntry(pe);
				nextPe.x--;
				nextPe.path.add((byte)'L');
				nextPe.prev = pe;
				peList.add(nextPe);
			}
			if (pe.x < wx-1 && last != 'L') {
				PathEntry nextPe = new PathEntry(pe);
				nextPe.x++;
				nextPe.path.add((byte)'R');
				nextPe.prev = pe;
				peList.add(nextPe);
			}
			if (0 < pe.y && last != 'D') {
				PathEntry nextPe = new PathEntry(pe);
				nextPe.y--;
				nextPe.path.add((byte)'U');
				nextPe.prev = pe;
				peList.add(nextPe);
			}
			if (pe.y < wy-1 && last != 'U') {
				PathEntry nextPe = new PathEntry(pe);
				nextPe.y++;
				nextPe.path.add((byte)'D');
				nextPe.prev = pe;
				peList.add(nextPe);
			}
		}
		throw new SolverException("Program error or bad data.");
	}
	
	public World solve() {
		boolean[] mask = new boolean[firstWorld.getTile().length];
		World world = firstWorld;
		for (SolveStep solveStep : solveStepList) {
			List<Integer> groupIdxs = solveStep.groupIdxs;
			boolean[] mask2 = new boolean[mask.length];
			System.arraycopy(mask, 0, mask2, 0, mask2.length);
			
			if (solveStep.stepType == 0) {
				world = step(world, solveStep, groupIdxs, mask, mask2);
			} else if (solveStep.stepType == 1) {
				World nextWorld1 = step(world, solveStep, groupIdxs, mask, mask2);
				Collections.reverse(groupIdxs);
				World nextWorld2 = step(world, solveStep, groupIdxs, mask, mask2);
				if (nextWorld1.getExPath().length < nextWorld2.getExPath().length) {
					world = nextWorld1;
				} else {
					world = nextWorld2;
				}
			} else if (solveStep.stepType == 2) {
				for (int groupIdx : solveStep.groupIdxs) {
					world = moveZero(world, mask, groupIdx%wx, groupIdx/wx);
				}
			}
			
			mask = mask2;
		}
		for (int i=fakeStartPath.path.size()-1;i>=0;i--) {
			char c = 0;
			switch (fakeStartPath.path.get(i)) {
			case 'U':
				c = 'D';
				break;
			case 'D':
				c = 'U';
				break;
			case 'L':
				c = 'R';
				break;
			case 'R':
				c = 'L';
				break;
			}
			world = world.getNext((byte) wx, (byte) wy, (byte) c);
		}
				
		return world;
	}
	private World step(World world, SolveStep solveStep, List<Integer> groupIdxs, boolean[] mask, boolean[] mask2) {
		if (solveStep.stepType == 0) {
			// 最後の仕上げ。2x2になっているので整える
			{	// 左上(?)に目的の値を持っていく
				int toIdx = groupIdxs.get(2);
				byte targetNo = correctAnswer[groupIdxs.get(2)];
				int fromIdx = seekNoIdx(world, targetNo);
				world = movePanel(world, mask, fromIdx%wx, fromIdx/wx, toIdx%wx, toIdx/wx);
			}
			{	// 0 をゴール地点に移動する
				int firstIdx = groupIdxs.get(0);
				world = moveZero(world, mask, firstIdx%wx, firstIdx/wx);
			}
		} else {
			// 重み付けした順に埋めていく
			for (int groupIdx : groupIdxs) {
				mask2[groupIdx] = true;
			}
			int frontIdx = seekFrontIdx(world, groupIdxs.get(0), -1);
			int tailIdx = seekFrontIdx(world, groupIdxs.get(groupIdxs.size()-1), frontIdx);
			if (tailIdx == -1) {
				tailIdx = seekFrontIdx(world, groupIdxs.get(groupIdxs.size()-1), -1);
				frontIdx = seekFrontIdx(world, groupIdxs.get(0), tailIdx);
				if (frontIdx == -1) {
					frontIdx = tailIdx; 
				}
			}
			if (groupIdxs.size() > 1) {
				boolean firstFlag = true;
				for (int groupIdx : groupIdxs) {
					byte targetNo = correctAnswer[groupIdx];
					
//					if (targetNo == 27) {
//						targetNo = 27;
//					}
//					
					int targetIdx = seekNoIdx(world, targetNo);
					if (groupIdxs.get(groupIdxs.size()-1) == targetIdx) {
						// FIXME ものすごい大雑把な手数削減
						boolean[] maskT = new boolean[mask.length];
						System.arraycopy(mask, 0, maskT, 0, mask.length);
						maskT[targetIdx] = true;
						world = moveZero(world, maskT, frontIdx%wx, frontIdx/wx);
						firstFlag = false;
						continue;
					}
					if (!firstFlag) {
						{	// pushback(既に中に含まれていた場合は押し出す)
							boolean needPushBack = false;
							int n = 0;
							for (Integer groupIdx2 : groupIdxs) {
								n++;
								if (targetIdx == groupIdx2) {
									needPushBack = true;
									break;
								}
							}
							if (needPushBack) {
								// 今から挿入する先に含まれているので叩き出す
								//n = groupIdxs.size() - n;
								for (int i=0;i<n;i++) {
									world = moveZero(world, mask2, frontIdx%wx, frontIdx/wx);
									for (int j=0;j<groupIdxs.size();j++) {
										Integer groupIdx2 = groupIdxs.get(j);
										world = moveZero(world, mask, groupIdx2%wx, groupIdx2/wx);
									}
									world = moveZero(world, mask2, tailIdx%wx, tailIdx/wx);
								}
								// 円環の理から叩き出す
								int tIdx;
								{
									boolean[] mask2T = new boolean[mask2.length];
									System.arraycopy(mask2, 0, mask2T, 0, mask2.length);
									List<PathEntry> peList = findMinPath(mask2, frontIdx%wx, frontIdx/wx, tailIdx%wx, tailIdx/wx);;
									for (int i=0;i<peList.size();i++) {
										PathEntry pe = peList.get(i);
										mask2T[pe.x + wx*pe.y] = true;
									}
									tIdx = findSafeIdx(world, mask2T, frontIdx%wx, frontIdx/wx);
								}
								world = movePanel(world, mask2, frontIdx%wx, frontIdx/wx, tIdx%wx, tIdx/wx);
								// 叩き出したものを元に戻す
								boolean[] mask2T = new boolean[mask2.length];
								System.arraycopy(mask2, 0, mask2T, 0, mask2.length);
								mask2T[tIdx] = true;
								for (int i=0;i<n;i++) {
									world = moveZero(world, mask2T, tailIdx%wx, tailIdx/wx);
									for (int j=groupIdxs.size()-1;j>=0;j--) {
										Integer groupIdx2 = groupIdxs.get(j);
										world = moveZero(world, mask, groupIdx2%wx, groupIdx2/wx);
									}
									world = moveZero(world, mask2T, frontIdx%wx, frontIdx/wx);
								}
								targetIdx = seekNoIdx(world, targetNo);
							}
						}
						{	// moveFront
							// FIXME mask2ではなく、それらしくフィルタリングされたものを使うこと
							world = movePanel(world, mask2, targetIdx%wx, targetIdx/wx, tailIdx%wx, tailIdx/wx);
						}
					} else {
						{	// moveFront
							world = movePanel(world, mask, targetIdx%wx, targetIdx/wx, tailIdx%wx, tailIdx/wx);
						}
					}
					{	// PushIn
						boolean[] maskT2 = new boolean[mask.length];
						if (!firstFlag) {
							System.arraycopy(mask2, 0, maskT2, 0, maskT2.length);
							maskT2[tailIdx] = true;
							world = moveZero(world, maskT2, frontIdx%wx, frontIdx/wx);
							for (int groupIdx2 : groupIdxs) {
								world = moveZero(world, mask, groupIdx2%wx, groupIdx2/wx);
							}
						} else {
							System.arraycopy(mask, 0, maskT2, 0, maskT2.length);
							maskT2[tailIdx] = true;
							int tIdx = groupIdxs.get(groupIdxs.size() - 1);
							world = moveZero(world, maskT2, tIdx%wx, tIdx/wx);
						}
						world = moveZero(world, mask, tailIdx%wx, tailIdx/wx);
					}
					firstFlag = false;
				}
			} else {
				int groupIdx = groupIdxs.get(0);
				byte targetNo = correctAnswer[groupIdx];
				
				int targetIdx = 0;
				{	// ターゲットの番号の在処を探す
					for (int i=0;i<world.getTile().length;i++) {
						if (targetNo == world.getTile()[i]) {
							targetIdx = i;
							break;
						}
					}
				}
				world = movePanel(world, mask, targetIdx%wx, targetIdx/wx, groupIdx%wx, groupIdx/wx);
			}
		}
		return world;
	}
	private World movePanel(World world, boolean[] mask, int fromX, int fromY, int toX, int toY) {
		byte targetNo = world.getTile()[fromX+fromY*world.getWx()];
		List<PathEntry> peList = findMinPath(mask, fromX, fromY, toX, toY);
		PathEntry lastPe = peList.get(0);
		for (int i=1;i<peList.size();i++) {
			PathEntry pe = peList.get(i);
			boolean[] mask2 = new boolean[mask.length];
			System.arraycopy(mask, 0, mask2, 0, mask2.length);
			mask2[lastPe.x + lastPe.y*wx] = true;
			
			world = moveZero(world, mask2, pe.x, pe.y);
			world = world.getNext((byte) wx, (byte) wy, seekNear(world, targetNo, world.getCursorX(), world.getCursorY()));
			
			lastPe = pe;
		}
		if (world == null) {
			throw new SolverException("Program error or bad data.");
		}
		world.setExPath(SolverUtil.compact(world.getExPath()));
		return world;
	}
	private World moveZero(World world, boolean[] mask, int x, int y) {
		List<PathEntry> peList = findMinPath(mask, world.getCursorX(), world.getCursorY(), x, y);
		for (byte d:peList.get(peList.size()-1).path) {
			world = world.getNext((byte)wx, (byte)wy, d);
		}
		return world;
	}
	private int seekNoIdx(World world, int targetNo) {
		int targetIdx = -1;
		for (int i=0;i<world.getTile().length;i++) {
			if (targetNo == world.getTile()[i]) {
				targetIdx = i;
				break;
			}
		}
		return targetIdx;
	}
	private int findSafeIdx(World world, boolean[] mask, int x, int y) {
		List<int[]> posList = new LinkedList<int[]>();
		posList.add(new int[]{x,y});
		boolean[] checked = new boolean[weights.length];
		while (posList.size() > 0) {
			int[] pos = posList.remove(0);
			int idx = pos[0] + wx * pos[1];
			if (pos[0] < 0 || wx <= pos[0] || pos[1] < 0 || wy <= pos[1]) {
				continue;
			}
			if (checked[idx]) {
				continue;
			}
			checked[idx] = true;
			if (!mask[idx] && weights[idx] != -1) {
				return idx;
			}
			posList.add(new int[]{pos[0],pos[1]-1});
			posList.add(new int[]{pos[0]-1,pos[1]});
			posList.add(new int[]{pos[0],pos[1]+1});
			posList.add(new int[]{pos[0]+1,pos[1]});
		}
		return 0;
	}
	private int seekFrontIdx(World world, int idx, int exclude) {
		int x = idx%wx;
		int y = idx/wx;
		byte target = weights[x+wx*y];
		byte[] ts = weights;
		if (0<x && ts[(x-1)+wx*y] != -1 && ts[(x-1)+wx*y] < target) {
			if (exclude != (x-1)+wx*y) {
				return (x-1)+wx*y;
			}
		}
		if (x<wx-1 && ts[(x+1)+wx*y] != -1 && ts[(x+1)+wx*y] < target) {
			if (exclude != (x+1)+wx*y) {
				return (x+1)+wx*y;
			}
		}
		if (0<y && ts[x+wx*(y-1)] != -1 && ts[x+wx*(y-1)] < target) {
			if (exclude != x+wx*(y-1)) {
				return x+wx*(y-1);
			}
		}
		if (y<wy-1 && ts[x+wx*(y+1)] != -1 && ts[x+wx*(y+1)] < target) {
			if (exclude != x+wx*(y+1)) {
				return x+wx*(y+1);
			}
		}
		return -1;
	}
	private byte seekNear(World world, byte target, int x, int y) {
		if (0<x && world.getTile()[(x-1)+wx*y] == target) {
			return 'L';
		}
		if (x<wx-1 && world.getTile()[(x+1)+wx*y] == target) {
			return 'R';
		}
		if (0<y && world.getTile()[x+wx*(y-1)] == target) {
			return 'U';
		}
		if (y<wy-1 && world.getTile()[x+wx*(y+1)] == target) {
			return 'D';
		}
		return 0;
	}
	
	public List<SolveStep> getSolveStepList() {
		return solveStepList;
	}
	public int getSolverType() {
		return 4;
	}
}
