package net.cattaka.slidepuzzle.solver;

public class World {
	private int hashCode;
	private byte[] tile = new byte[36];
	private byte wx;
	private byte wy;
	private byte cursorX;
	private byte cursorY;
	private byte[] exPath;
	private int exScore;
	private int exScore2;

	public World(byte[] tile, byte cursorX, byte cursorY, byte wx, byte wy) {
		super();
		this.tile = tile;
		this.wx = wx;
		this.wy = wy;
		
		this.cursorX = cursorX;
		this.cursorY = cursorY;
		for (int i = 0; i < tile.length; i++) {
			if (tile[i] == 0) {
				this.cursorX = (byte) (i % wx);
				this.cursorY = (byte) (i / wx);
			}
		}

		recalcHash();
	}

	public static World createWorld(byte[] tile, byte wx, byte wy) {
		byte cursorX = -1;
		byte cursorY = -1;
		for (int i = 0; i < tile.length; i++) {
			if (tile[i] == 0) {
				cursorX = (byte) (i % wx);
				cursorY = (byte) (i / wx);
			}
		}
		if (cursorX == -1 || cursorY == -1) {
			throw new RuntimeException();
		}
		World world = new World(tile, cursorX, cursorY, wx, wy);
		world.setExPath(new byte[0]);
		return world;
	}
	public static World createWorld(String dat, int wx, int wy) {
		byte[] tile = new byte[dat.length()];
		for (int i = 0; i < dat.length(); i++) {
			if (dat.charAt(i) == '=') {
				tile[i] = -1;
			} else {
				tile[i] = (byte) Integer.parseInt(
						String.valueOf(dat.charAt(i)), 36);
			}
		}

		return World.createWorld(tile, (byte) wx, (byte) wy);
	}

	private void recalcHash() {
		hashCode = 0;
		for (int i = 0; i < tile.length; i++) {
			hashCode += (int) tile[i] * (7 << i);
		}
	}

	public byte[] getTile() {
		return tile;
	}

	public World getNext(byte wx, byte wy, byte d) {
		byte[] nextTile = new byte[tile.length];
		byte nextCursorX;
		byte nextCursorY;

		System.arraycopy(tile, 0, nextTile, 0, nextTile.length);
		nextCursorX = cursorX;
		nextCursorY = cursorY;

		switch (d) {
		case 'U':
			nextCursorY--;
			if (nextCursorY < 0) {
				return null;
			}
			break;
		case 'D':
			nextCursorY++;
			if (nextCursorY >= wy) {
				return null;
			}
			break;
		case 'L':
			nextCursorX--;
			if (nextCursorX < 0) {
				return null;
			}
			break;
		case 'R':
			nextCursorX++;
			if (nextCursorX >= wx) {
				return null;
			}
			break;
		default:
			return null;
		}
		byte targetVal = nextTile[nextCursorX + wx * nextCursorY];
		if (targetVal == (byte) -1) {
			return null;
		}
		nextTile[nextCursorX + wx * nextCursorY] = tile[cursorX + wx
				* cursorY];
		nextTile[cursorX + wx * cursorY] = targetVal;

		World nextWorld = new World(nextTile, nextCursorX, nextCursorY, wx,
				wy);
		byte[] newPath = new byte[exPath.length + 1];
		System.arraycopy(exPath, 0, newPath, 0, exPath.length);
		newPath[exPath.length] = (byte) d;
		nextWorld.setExPath(newPath);
		return nextWorld;
	}

	public Byte getAnotherWay(byte last, int x, int y, int wx, int wy) {
		int count = 0;
		byte c = 'L';
		if (last != 'R' && !isWall(tile, x - 1, y, wx, wy)) {
			count++;
			c = 'L';
		}
		if (last != 'L' && !isWall(tile, x + 1, y, wx, wy)) {
			count++;
			c = 'R';
		}
		if (last != 'D' && !isWall(tile, x, y - 1, wx, wy)) {
			count++;
			c = 'U';
		}
		if (last != 'U' && !isWall(tile, x, y + 1, wx, wy)) {
			count++;
			c = 'D';
		}
		return (count == 1) ? c : null;
	}

	private boolean isWall(byte[] tile, int x, int y, int wx, int wy) {
		if (x < 0 || wx <= x || y < 0 || wy <= y) {
			return true;
		} else if (tile[x + y * wx] == -1) {
			return true;
		}
		return false;
	}

	@Override
	public boolean equals(Object obj) {
		byte[] ot = ((World) obj).getTile();
		for (int i = 0; i < tile.length; i++) {
			if (tile[i] != ot[i]) {
				return false;
			}
		}
		return true;
	}

	@Override
	public int hashCode() {
		return hashCode;
	}

	public byte[] getExPath() {
		return exPath;
	}

	public void setExPath(byte[] exPath) {
		this.exPath = exPath;
	}

	public int getExScore() {
		return exScore;
	}

	public void setExScore(int exScore) {
		this.exScore = exScore;
	}

	public int getExScore2() {
		return exScore2;
	}

	public void setExScore2(int exScore2) {
		this.exScore2 = exScore2;
	}

	public byte getCursorX() {
		return cursorX;
	}

	public byte getCursorY() {
		return cursorY;
	}

	public byte getWx() {
		return wx;
	}

	public byte getWy() {
		return wy;
	}

	public void setWx(byte wx) {
		this.wx = wx;
	}

	public void setWy(byte wy) {
		this.wy = wy;
	}

	public void setCursorX(byte cursorX) {
		this.cursorX = cursorX;
	}

	public void setCursorY(byte cursorY) {
		this.cursorY = cursorY;
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (int y = 0; y < wy; y++) {
			for (int x = 0; x < wx; x++) {
				if (tile[x + wx * y] == -1) {
					sb.append(" ==");
				} else if (tile[x + wx * y] == x + wx * y + 1) {
					sb.append(String.format("+%2d", tile[x + wx * y]));
				} else {
					sb.append(String.format(" %2d", tile[x + wx * y]));
				}
			}
			sb.append('\n');
		}
		return sb.toString();
	}
}
